from collections import namedtuple

from trading_bots.bots import Bot
from trading_bots.contrib.clients import Market, Side
from trading_bots.contrib.clients import bitfinex, bitstamp, buda, kraken
from trading_bots.contrib.converters.open_exchange_rates import OpenExchangeRates
from trading_bots.utils import truncate_to


class SimpleLimit(Bot):
    label = 'SimpleLimit'
    market_clients = [
        bitfinex.BitfinexMarket,
        bitstamp.BitstampMarket,
        kraken.KrakenMarket,
    ]

    def _setup(self, config):
        # Set market
        self.market = Market(config['market'])
        # Init variables
        self.bid_price, self.ask_price = None, None
        self.base_amount, self.quote_amount = None, None
        # Set buda trading client
        self.buda = buda.BudaTrading(
            self.market, dry_run=self.dry_run, timeout=self.timeout, logger=self.log, store=self.store)
        # Set converter
        self.converter = OpenExchangeRates(timeout=self.timeout)
        # Set reference market client
        reference_config = config['reference']
        self.reference = self._get_market_client(reference_config['name'], reference_config['market'])
        assert self.reference.market.base == self.market.base

    def _algorithm(self):
        # Setup
        self.log.info(f'Preparing prices using {self.reference.name} {self.reference.market.code}')
        self.prepare_prices()
        # Cancel open orders
        self.log.info('Closing open orders')
        self.cancel_orders()
        # Get available balances
        self.prepare_amounts()
        # Start strategy
        self.log.info('Starting order deployment')
        # Deploy orders
        deploy_list = self.get_deploy_list()
        self.deploy_orders(deploy_list)

    def _abort(self):
        self.log.error('Aborting strategy, cancelling all orders')
        try:
            self.cancel_orders()
        except Exception:
            self.log.exception(f'Failed!, some orders might not be cancelled')
            raise
        else:
            self.log.info(f'All open orders were cancelled')

    def prepare_prices(self):
        self.log.info(f'Preparing prices')
        prices_config = self.config['prices']
        ref_bid, ref_ask = self._get_reference_prices()
        self.log.info(f'{self.market.base} reference prices on {self.reference.name}: Bid: {ref_bid} Ask: {ref_ask}')
        # Set offset prices from reference and price multiplier
        self.bid_price = ref_bid * prices_config['buy_multiplier']
        self.ask_price = ref_ask * prices_config['sell_multiplier']
        self.log.info(f'{self.market} calculated prices: Bid: {self.bid_price} Ask: {self.ask_price}')

    def prepare_amounts(self):
        self.log.info(f'Preparing amounts')
        # Get amounts from configs
        amounts_config = self.config['amounts']
        max_base = amounts_config['max_base']
        max_quote = amounts_config['max_quote']
        # Set final bid and ask amounts
        self.base_amount = min(max_base, self.buda.wallets.base.get_available())
        self.quote_amount = min(max_quote, self.buda.wallets.quote.get_available())
        self.log.debug(' | '.join([
            'Amounts',
            f'Bid: {self.quote_amount} {self.market.quote}',
            f'Ask: {self.base_amount} {self.market.base}'
        ]))

    def get_deploy_list(self):

        Order = namedtuple('order', 'amount price side')
        deploy_list = []

        # Available is on quote currency when side is sell
        def quote_to_base_amount(_amount, _price):
            quote_amount = _amount / _price
            return self.truncate_amount(quote_amount)

        buy_order_price = self.truncate_price(self.bid_price)
        buy_order_amount = quote_to_base_amount(self.quote_amount, buy_order_price)
        sell_order_price = self.truncate_price(self.ask_price)
        sell_order_amount = self.truncate_amount(self.base_amount)

        if buy_order_amount > self.buda.min_order_amount:
            deploy_list.append(Order(buy_order_amount, buy_order_price, Side.BUY))
        if sell_order_amount > self.buda.min_order_amount:
            deploy_list.append(Order(sell_order_amount, sell_order_price, Side.SELL))
        return deploy_list

    def deploy_orders(self, deploy_list: list):
        self.log.info(f'Deploying {len(deploy_list)} new orders')
        for order in deploy_list:
            self.buda.place_limit_order(
                side=order.side,
                amount=order.amount,
                price=order.price,
            )

    def cancel_orders(self, remove_list: list=None):
        remove_list = remove_list or []
        if len(remove_list) > 0:
            self.log.info(f'Canceling {len(remove_list)} orders')
        else:
            self.log.info(f'Canceling open orders')
        self.buda.cancel_orders(remove_list)

    def _get_reference_prices(self):
        ref_bid, ref_ask = self.reference.get_spread_details()
        # Convert reference_price if reference market differs from current market
        if self.reference.market != self.market:
            # Get conversion rate (eg CLP/USD from OpenExchangeRates)
            rate = self.converter.get_rate_for(self.reference.market.quote, self.market.quote)
            self.log.info(f'{self.reference.market.quote}/{self.market.quote} rate: {rate} from {self.converter.name}')
            # Get market price according to reference (eg BTC/CLP converted from converter's BTC/USD)
            ref_bid *= rate
            ref_ask *= rate
        return ref_ask, ref_ask

    def _get_market_client(self, name, market):
        for client in self.market_clients:
            if client.name == name:
                return client(market, client=None, dry_run=self.dry_run,
                              timeout=self.timeout, logger=self.log, store=self.store)
        raise NotImplementedError(f'Client {name} not found!')

    def truncate_amount(self, value):
        return truncate_to(value, self.market.base)

    def truncate_price(self, value):
        return truncate_to(value, self.market.quote)



import time

from trading_bots.bots import Bot


class Example(Bot):
    label = 'Example'

    def _setup(self, config):
        self.work_time = config['work_time']

    def _algorithm(self):
        self.log.info('This is an example bot')
        self.log.info(f'Doing work for {self.work_time} seconds...')
        time.sleep(self.work_time)
        self.log.info('Finished!')

    def _abort(self):
        self.log.warning('Exit!')



import codecs
import importlib
import re

import yaml


def load_class_by_name(name: str):
    """Given a dotted path, returns the class"""
    mod_path, _, cls_name = name.rpartition('.')
    mod = importlib.import_module(mod_path)
    cls = getattr(mod, cls_name)
    return cls


def load_yaml_file(file_path: str):
    """Load a YAML file from path"""
    with codecs.open(file_path, 'r') as f:
        return yaml.safe_load(f)



"""
Global Trading-Bots exception and warning classes.
"""


class AppRegistryNotReady(Exception):
    """The trading_bots.bots registry is not populated yet"""
    pass


class ImproperlyConfigured(Exception):
    """Trading-Bots is somehow improperly configured"""
    pass



from datetime import datetime
from time import sleep

from trading_bots.bots import Bot
from trading_bots.contrib.clients import Market, Side
from trading_bots.contrib.clients import buda
from trading_bots.utils import truncate_to


class AnyToAny(Bot):
    label = 'AnyToAny'

    def _setup(self, config):
        # Get configs
        self.from_currency = config['from']['currency']
        self.from_address = config['from']['address']
        self.to_currency = config['to']['currency']
        self.to_withdraw = config['to']['withdraw']
        self.to_address = config['to']['address']
        # Set market
        self.market = self._get_market(self.from_currency, self.to_currency)
        # Set side
        self.side = Side.SELL if self.market.base == self.from_currency else Side.BUY
        # Set buda trading client
        self.buda = buda.BudaTrading(
            self.market, dry_run=self.dry_run, timeout=self.timeout, logger=self.log, store=self.store)
        # Get deposits
        self.deposits = self.store.get(self.from_currency + '_deposits') or {}
        # Set start date
        self.start_date = datetime.utcnow()

    def _algorithm(self):
        # Get new deposits
        self.log.info(f'Checking for new {self.from_currency} deposits')
        self.update_deposits()
        # Convert pending amounts
        self.log.info('Converting pending amounts')
        self.process_conversions()
        # Get available balances
        self.log.info('Processing pending withdrawals')
        self.process_withdrawals()

    def _abort(self):
        pass

    def update_deposits(self):
        # Get deposits
        deposits = self.deposits
        # Set wallet from relevant currency according to side
        from_wallet = self.buda.wallets.quote if self.side == Side.BUY else self.buda.wallets.base
        # Get and filter deposits
        new_deposits = from_wallet.get_deposits()
        if self.from_address != 'Any':
            new_deposits = [deposit for deposit in new_deposits if deposit.data.address == self.from_address]
        new_deposits = [deposit for deposit in new_deposits if deposit.created_at >= self.start_date]
        # Update states on existing keys and add new keys with base structure
        for deposit in new_deposits:
            idx = str(deposit.id)
            if idx in deposits.keys():
                if deposit.state != deposits[idx]['state']:
                    deposits[idx]['state'] = deposit.state
            else:
                deposits[idx] = {
                    'state': deposit.state,
                    'amounts': {'original_amount': deposit.amount.amount,
                                'converted_amount': 0,
                                'converted_value': 0},
                    'orders': [],
                    'pending_withdrawal': self.to_withdraw
                    }
            self.store.store(self.from_currency + '_deposits', deposits)
            self.deposits = deposits

    def process_conversions(self):
        # Get deposits
        deposits = self.deposits
        for deposit_id in deposits:
            # Calculate remaining amount to convert
            original_amount = deposits[deposit_id]['amounts']['original_amount']
            converted_amount = deposits[deposit_id]['amounts']['converted_amount']
            converted_value = deposits[deposit_id]['amounts']['converted_value']
            remaining = original_amount - converted_amount
            if deposits[deposit_id]['state'] == 'confirmed' and remaining > 0:
                if self.side == Side.BUY:  # Change amount to base currency for order creation purposes
                    remaining = self.buda.client.quotation_market(self.market.base + '-' + self.market.quote,
                                                                  'bid_given_spent_quote',
                                                                  remaining).order_amount.amount
                remaining = truncate_to(remaining, self.market.base)
                # Convert remaining amount using market order
                order = self.buda.place_market_order(self.side, remaining)
                # Wait for traded state to set updated values
                if order:
                    self.log.info(f'{self.side} market order placed, waiting for traded state')
                    while order.state != 'traded':
                        order = self.buda.client.order_details(order.id)
                        sleep(1)
                    self.log.info(f'{self.side} order traded, updating store values')
                    converted_amount += order.total_exchanged.amount if self.side == Side.BUY\
                                                                and order.state == 'traded'\
                                                                else order.traded_amount.amount
                    converted_value += order.traded_amount.amount if self.side == Side.BUY\
                                                                and order.state == 'traded'\
                                                                else order.total_exchanged.amount
                    converted_value -= order.paid_fee.amount  # Fee deducted so it wont interfere with withdrawal
                    deposits[deposit_id]['orders'].append(order.id)  # Save related orders for debugging
                # Save new values
                deposits[deposit_id]['amounts']['converted_amount'] = converted_amount
                deposits[deposit_id]['amounts']['converted_value'] = converted_value
                self.store.store(self.from_currency + '_deposits', deposits)
                self.deposits = deposits

    def process_withdrawals(self):
        # Get deposits
        deposits = self.deposits
        # Set wallet from relevant currency according to side
        to_wallet = self.buda.wallets.base if self.side == Side.BUY else self.buda.wallets.quote
        for deposit_id in deposits:
            # Filter deposits already converted and pending withdrawal
            if deposits[deposit_id]['state'] == 'confirmed' \
                    and deposits[deposit_id]['pending_withdrawal'] \
                    and deposits[deposit_id]['amounts']['original_amount'] == deposits[deposit_id]['amounts']['converted_amount']:
                withdrawal_amount = deposits[deposit_id]['amounts']['converted_value']
                withdrawal_amount = truncate_to(withdrawal_amount, self.to_currency)
                available = to_wallet.get_available()
                if withdrawal_amount <= available:  # We cannot withdraw more than available balance
                    w = to_wallet.request_withdrawal(withdrawal_amount, self.to_address, subtract_fee=True)
                    if w.state == 'pending_preparation':  # Check state to set and store updated values
                        self.log.info(f'{self.to_currency} withdrawal request received, updating store values')
                        deposits[deposit_id]['pending_withdrawal'] = False
                        self.store.store(self.from_currency + '_deposits', deposits)
                        self.deposits = deposits
                    else:
                        self.log.warning('Withdrawal failed')
                else:
                    self.log.warning(f'Available balance not enough for withdrawal amount {amount} {self.to_currency}')

    def _get_market(self, from_currency, to_currency):
        public_client = buda.BudaPublic()
        buda_markets = public_client.client.markets()
        bases = [market.base_currency for market in buda_markets]
        quotes = [market.quote_currency for market in buda_markets]

        if from_currency in bases and to_currency in quotes:
            market = Market((from_currency, to_currency))
        elif from_currency in quotes and to_currency in bases:
            market = Market((to_currency, from_currency))
        else:
            raise NotImplementedError(f'No compatible market found!')
        return market



from trading_api_wrappers import Kraken

from .base import *

__all__ = [
    'KrakenPublic',
    'KrakenAuth',
    'KrakenMarket',
    'KrakenWallet',
    'KrakenTrading',
]


class KrakenPublic(BaseClient):
    name = 'Kraken'

    def _client(self):
        return Kraken.Public(timeout=self.timeout)


class KrakenAuth(BaseClient):
    name = 'Kraken'

    def _client(self):
        key = self.credentials['key']
        secret = self.credentials['secret']
        return Kraken.Auth(key, secret, timeout=self.timeout)


class KrakenMarket(MarketClient, KrakenPublic):

    def _market_id(self):
        return str(self.market).replace('BTC', 'XBT')

    def _ticker(self):
        result = self.client.ticker(symbol=self.market_id)['result']
        # Get result first key
        return result[next(iter(result))]

    def _order_book(self, side: Side=None):
        result = self.client.order_book(symbol=self.market_id)['result']
        # Get result first key
        order_book = result[next(iter(result))]
        bids, asks = order_book['bids'], order_book['asks']
        if side:
            return bids if side == Side.BUY else asks
        return OrderBook(bids=bids, asks=asks)

    def _order_book_entry_amount(self, order):
        return float(order[1])

    def _order_book_entry_price(self, order):
        return float(order[0])


class KrakenWallet(WalletClient, KrakenAuth):
    withdrawal_fees = {
        'BCH': 0.0005,
        'BTC': 0.0005,
        'ETH': 0.01,
        'LTC': 0.01,
    }

    def _balance(self, currency, available_only=False):
        # TODO: How to get available_only?
        if available_only:
            self.log.warning('available_only option is not implemented!')
        asset = currency.replace('BTC', 'XXBT').replace('ETH', 'XETH').replace('XLM', 'XXLM').replace('USD', 'ZUSD')
        balance = self.client.balance()
        return float(balance['result'][asset])

    @staticmethod
    def _filter_state(items, state: str=None):
        if state:
            items = [i for i in items if i['status'] == state]
        return items

    def _deposits(self, currency: str):
        mapping = {
            'BCH': ('BCH', 'Bitcoin Cash'),
            'BTC': ('XBT', 'Bitcoin'),
            'ETH': ('XETH', 'Ether (Hex)'),
            'LTC': ('LTC', 'Litecoin'),
        }
        asset, method = mapping[currency]
        return self.client.deposit_status(asset, method)['result']

    def _withdrawals(self, currency: str):
        mapping = {
            'BCH': ('BCH', 'Bitcoin Cash'),
            'BTC': ('XBT', 'Bitcoin'),
            'ETH': ('XETH', 'Ether'),
            'LTC': ('LTC', 'Litecoin'),
        }
        asset, method = mapping[currency]
        return self.client.withdraw_status(asset, method)['result']

    def _withdraw(self, currency: str, amount: float, address: str, subtract_fee: bool=False):
        asset = currency.replace('BTC', 'XBT').replace('ETH', 'XETH')
        return self.client.withdraw(asset, amount, address)


class KrakenTrading(TradingClient, KrakenAuth, KrakenMarket):
    wallet_client = KrakenWallet
    has_margin_trading = True
    min_order_amount_mapping = {
        'BCH': 0.002,
        'BTC': 0.002,
        'ETH': 0.02,
        'LTC': 0.002,
    }

    def _open_orders(self):
        return self.client.open_orders()['result']['open'].values()

    def _order_amount(self, order):
        return order['remaining_amount']

    def _cancel_order(self, order):
        return self.client.cancel_order(order['id'])

    def _order_details(self, order_id: int):
        return self.client.query_orders([order_id])['result'][order_id]

    def _place_order(self, side: Side, o_type: OrderType, amount: float, price: float=None):
        return self.client.add_order(self.market_id, side.value, o_type.value, amount, price)

    def _position_amount(self, position):
        return float(position['vol'])

    def _open_positions(self):
        positions = self.client.open_positions()['result'].values()
        return [p for p in positions if p['pair'] == self.market_id and p['posstatus'] == 'open']

    def _open_position(self, side: Side, p_type: OrderType, amount: float, price: float=None, leverage: float=None):
        return self.client.add_order(self.market_id, side.value, p_type.value, amount, price, leverage=leverage)



import math
from datetime import datetime

DECIMALS = {
    # Fiat
    'ARS': 2,
    'BRL': 2,
    'CLP': 2,
    'COP': 2,
    'EUR': 2,
    'PEN': 2,
    'USD': 2,
    # Crypto
    'BCH': 8,
    'BTC': 8,
    'ETH': 8,
    'LTC': 8,
}


def get_iso_time_str(timestamp=None):
    """Get the ISO time string from a timestamp or date obj. Returns current time str if no timestamp is passed"""
    if isinstance(timestamp, (int, float)):
        timestamp = datetime.utcfromtimestamp(timestamp)
    if timestamp is None:
        timestamp = datetime.utcnow()
    return timestamp.isoformat(sep=' ', timespec='seconds')


def truncate(value: float, decimal_places: int):
    """Truncates a value to a number of decimals places"""
    return math.trunc(value * (10 ** decimal_places)) / (10 ** decimal_places)


def truncate_to(value: float, currency: str):
    """Truncates a value to the number of decimals corresponding to the currency"""
    decimal_places = DECIMALS.get(currency.upper(), 2)
    return truncate(value, decimal_places)


def spread_value(value: float, spread_p: float):
    """Returns a lower and upper value separated by a spread percentage"""
    upper = value * (1 + spread_p)
    lower = value / (1 + spread_p)
    return lower, upper


def validate(name: str, value, condition: bool):
    """Validates value on condition"""
    assert condition, f'{name} is invalid! ({name}: {value})'


def validate_age(name: str, tolerance, from_timestamp, to_timestamp):
    """Check if age is valid (within tolerance)"""
    age = to_timestamp - from_timestamp
    assert age <= tolerance, f'{name} is too old! (Age: {age} > Tolerance: {tolerance})'



from trading_bots.bots import Bot
from trading_bots.contrib.clients import Market, Side
from trading_bots.contrib.clients import bitstamp, buda
from trading_bots.utils import truncate_to, get_iso_time_str
import pandas as pd
import talib
import time
from operator import itemgetter


class TechnicalAnalysis(Bot):
    label = 'TechnicalAnalysis'
    market_clients = [
        bitstamp.BitstampMarket,
    ]

    def _setup(self, config):
        # Set market
        self.market = Market(config['market'])
        # Init variables
        self.base_amount, self.quote_amount = config['amounts']['max_base'],  config['amounts']['max_quote']
        self.store_keys = ('trades', self.market.code.lower())
        self.position = self.store.get('position') or dict(status='closed')
        # Set talib configs
        talib_config = config['talib']
        self.bbands_periods = talib_config['bbands']['periods']
        self.rsi_periods = talib_config['rsi']['periods']
        self.rsi_overbought = talib_config['rsi']['overbought']
        self.rsi_oversold = talib_config['rsi']['oversold']
        # Set buda trading client
        self.buda = buda.BudaTrading(
            self.market, dry_run=self.dry_run, timeout=self.timeout, logger=self.log, store=self.store)
        # Set reference market client
        reference_config = config['reference']
        self.candle_interval = config['reference']['candle_interval']
        self.reference = self._get_market_client(reference_config['name'], reference_config['market'])
        assert self.reference.market.base == self.market.base

    def _algorithm(self):
        # Update candle data and TA indicators
        self.log.info(f'Getting trades from {self.reference.name} {self.reference.market.code}')
        from_time = time.time() - 60*60*24
        trades = self.get_trades(from_time)
        # Create pandas DataFrame from trades and set date as index
        df = pd.DataFrame(trades)
        df.index = df.timestamp.apply(lambda x: pd.datetime.utcfromtimestamp(x))
        # Build 5min candles from trades DataFrame [open, high, close, low]
        df = df.rate.resample(self.candle_interval).ohlc()
        # Calculate Bolliger Bands and RSI from talib
        df['bb_lower'], df['bb_middle'], df['bb_upper'] = talib.BBANDS(df.close, timeperiod=self.bbands_periods)
        df['rsi'] = talib.RSI(df.close, timeperiod=self.rsi_periods)
        lower = self.truncate_price(df.bb_lower[-1])
        middle = self.truncate_price(df.bb_middle[-1])
        upper = self.truncate_price(df.bb_upper[-1])
        self.log.info(f'BB_lower: {lower} | BB_middle: {middle} | BB_upper: {upper}')
        self.log.info(f'RSI: {df.rsi[-1]:.2f}')
        # Check if our position is open or closed
        if self.position['status'] == 'closed':
            # Try conditions to open position
            self.log.info(f'Position is closed, checking to open')
            if df.close[-1] < df.bb_lower[-1] and df.rsi[-1] < self.rsi_oversold:
                # Last price is lower than the lower BBand and RSI is oversold, BUY!
                self.log.info(f'Market oversold! BUY!')
                amount = self.get_amount(Side.SELL)
                tx = self.buda.place_market_order(Side.SELL, amount)
                self.position = {
                    'status': 'open',
                    'side': Side.SELL.value,
                    'amount': tx.amount.amount
                }
            elif df.close[-1] > df.bb_upper[-1] and df.rsi[-1] > self.rsi_overbought:
                # Last price is higher than the upper BBand and RSI is overbought, SEll!
                self.log.info(f'Market overbought! SEll!')
                amount = self.get_amount(Side.BUY)
                tx = self.buda.place_market_order(Side.BUY, amount)
                self.position = {
                    'status': 'open',
                    'side': Side.BUY.value,
                    'amount': tx.amount.amount
                }
            else:
                self.log.info(f'Market conditions unmet to open position')
        else:
            self.log.info(f'Position is open, checking to close')
            if self.position['side'] == Side.BUY.value and df.rsi[-1] >= 30:
                # RSI is back to normal, close Buy position
                self.log.info(f'Market is back to normal, closing position')
                amount = self.position['amount']
                tx = self.buda.place_market_order(Side.SELL, amount)
                remaining = self.position['amount'] - tx.amount.amount
                if remaining == 0:
                    self.position = {'status': 'closed'}
                else:
                    self.position['amount'] = remaining
            elif self.position['side'] == Side.SELL.value and df.rsi[-1] <= 70:
                # RSI is back to normal, close Sell position
                self.log.info(f'Market is back to normal, closing position')
                amount = self.position['amount']
                tx = self.buda.place_market_order(Side.BUY, amount)
                remaining = self.position['amount'] - tx.amount.amount
                if remaining == 0:
                    self.position = {'status': 'closed'}
                else:
                    self.position['amount'] = remaining
            else:
                self.log.info(f'Market conditions unmet to close position')
        self.store.set('position', self.position)

    def _abort(self):
        pass

    def get_amount(self, side):
        if side == Side.BUY:
            return self.truncate_amount(self.base_amount)
        elif side == Side.SELL:
            price = self.buda.quote_price(side, self.quote_amount)
            return self.truncate_amount(self.quote_amount / price)

    def get_trades(self, from_timestamp: float):
        # Getting previous trades from store
        self.log.debug(f'Fetching trades since %s', get_iso_time_str(from_timestamp))
        prev_trades = self._get_previous_trades() or []
        # Check trades to establish timestamp
        if prev_trades:
            first_trade = prev_trades[0]['timestamp']
            last_trade = prev_trades[-1]['timestamp']
            self.log.debug(f'{len(prev_trades)} previous trades: (%s ... %s)',
                           get_iso_time_str(first_trade), get_iso_time_str(last_trade))
            # Check first_trade < last_trade on file
            assert first_trade < last_trade, f"File's first_trade > last_trade!)"
            # Use file's last trade timestamp whenever possible
            if first_trade < from_timestamp < last_trade:
                query_timestamp = last_trade
            else:
                query_timestamp = from_timestamp
                prev_trades = []
        else:
            self.log.info(f'No trades in store!')
            query_timestamp = from_timestamp
        # Add a null entry with n_days_timestamp (entries are bases on trades IDs)
        trades = [trade for trade in prev_trades if trade.get('timestamp') >= from_timestamp]
        trades.append(dict(timestamp=from_timestamp, rate=0.0, amount=0.0))
        # Poll trades from client
        trades_n = 1000
        n_calls = 0
        while trades_n == 1000:
            self.log.debug(f'Fetching {self.reference.name} trades since: %s',
                           get_iso_time_str(query_timestamp))
            if n_calls > 1:
                time.sleep(1)
            try:
                last_trades, query_timestamp = self._get_trades_call(query_timestamp)
            except Exception:
                self.log.exception(f'Failed obtaining {self.reference.name} trades!')
                raise
            n_calls += 1
            if last_trades:
                self.log.debug(f'Trades found: {len(last_trades)} (%s ... %s)',
                               get_iso_time_str(last_trades[0]['timestamp']),
                               get_iso_time_str(last_trades[-1]['timestamp']))
                trades.extend(last_trades)
            else:
                self.log.debug('No trades found')
            # Sort entries based on timestamp, then filter on n_days_timestamp
            trades.sort(key=itemgetter('timestamp'))
            # last trades count
            trades_n = len(last_trades)
            # Store trades
            self.log.debug(f'Storing {len(trades)} trades')
            self._store_trades(trades)
        return trades[1:]

    def reset_trades(self):
        self.store.hdel(*self.store_keys)

    def _get_trades_call(self, query_timestamp: float):
        one_day_seconds = 86400
        time_interval = 'day' if time.time() - query_timestamp >= one_day_seconds else 'hour'
        trades = self.reference.client.transactions(self.reference.market_id, time_interval)
        trades = [
            dict(timestamp=float(trade['date']),
                 rate=float(trade['price']),
                 amount=float(trade['amount']) * (1 if trade['type'] == '0' else -1))  # '0': buy, '1': sell
            for trade in trades
            if float(trade['date']) > query_timestamp]
        trades.sort(key=itemgetter('timestamp'))
        last_timestamp = trades[-1]['timestamp'] + 1 if trades else query_timestamp
        return trades, last_timestamp

    def _get_previous_trades(self):
        return self.store.hget(*self.store_keys, serializer='json')

    def _store_trades(self, trades: list):
        self.store.hset(*self.store_keys, value=trades, serializer='json')

    def _get_market_client(self, name, market):
        for client in self.market_clients:
            if client.name == name:
                print(client.name)
                return client(
                    market, client=None, dry_run=self.dry_run, timeout=self.timeout, logger=self.log,
                    store=self.store)
        raise NotImplementedError(f'Client {name} not found!')

    def truncate_amount(self, value):
        return truncate_to(value, self.market.base)

    def truncate_price(self, value):
        return truncate_to(value, self.market.quote)



from collections import namedtuple
from operator import attrgetter

import pytest

from trading_bots.conf import settings
from trading_bots.contrib import converters

CREDENTIALS = settings.credentials

CRYPTO_CONVERTERS = [
    converters.BitcoinAverage(),
    converters.CoinMarketCap(),
]

FIAT_CONVERTERS = [
    converters.CurrencyLayer(CREDENTIALS.get('CurrencyLayer')),
    converters.OpenExchangeRates(CREDENTIALS.get('OpenExchangeRates')),
]

Market = namedtuple('market', 'base quote id')

CRYPTO_MARKETS = [
    Market('BTC', 'CLP', 'upper'),
    Market('btc', 'CLP', 'lower_base'),
    Market('BTC', 'clp', 'lower_quote'),
    Market('BTC', 'BTC', 'same'),
]

FIAT_MARKETS = [
    Market('USD', 'CLP', 'upper'),
    Market('usd', 'CLP', 'lower_base'),
    Market('USD', 'clp', 'lower_quote'),
    Market('USD', 'BTC', 'same'),
]


@pytest.fixture(params=CRYPTO_CONVERTERS, ids=attrgetter('slug'))
def crypto_converter(request):
    return request.param


@pytest.fixture(params=FIAT_CONVERTERS, ids=attrgetter('slug'))
def fiat_converter(request):
    return request.param


@pytest.fixture(params=CRYPTO_MARKETS, ids=attrgetter('id'))
def crypto_market(request):
    return request.param


@pytest.fixture(params=FIAT_MARKETS, ids=attrgetter('id'))
def fiat_market(request):
    return request.param


def test_crypto_converter_instance(crypto_converter):
    assert isinstance(crypto_converter, converters.Converter)


def test_fiat_converter_instance(fiat_converter):
    assert isinstance(fiat_converter, converters.Converter)


def test_get_crypto_rate(crypto_converter, crypto_market):
    rate = crypto_converter.get_rate_for(crypto_market.base, to=crypto_market.quote)
    assert isinstance(rate, float)
    assert rate > 0


def test_get_fiat_rate(fiat_converter, fiat_market):
    rate = fiat_converter.get_rate_for(fiat_market.base, to=fiat_market.quote)
    assert isinstance(rate, float)
    assert rate > 0


def test_convert_crypto(crypto_converter, crypto_market):
    amount = crypto_converter.convert(0.5, currency=crypto_market.base, to=crypto_market.quote)
    assert isinstance(amount, float)
    assert amount > 0


def test_convert_fiat(fiat_converter, fiat_market):
    amount = fiat_converter.convert(0.5, currency=fiat_market.base, to=fiat_market.quote)
    assert isinstance(amount, float)
    assert amount > 0



from trading_api_wrappers import Bitstamp

from trading_bots.utils import truncate
from .base import *

__all__ = [
    'BitstampPublic',
    'BitstampAuth',
    'BitstampMarket',
    'BitstampWallet',
    'BitstampTrading',
]


class BitstampPublic(BaseClient):
    name = 'Bitstamp'

    def _client(self):
        return Bitstamp.Public(timeout=self.timeout)


class BitstampAuth(BaseClient):
    name = 'Bitstamp'

    def _client(self):
        key = self.credentials['key']
        secret = self.credentials['secret']
        customer_id = self.credentials['customer_id']
        return Bitstamp.Auth(key, secret, customer_id, timeout=self.timeout)


class BitstampMarket(MarketClient, BitstampPublic):

    def _market_id(self):
        return str(self.market).lower()

    def _ticker(self):
        return self.client.ticker(self.market_id)

    def _order_book(self, side: Side=None):
        order_book = self.client.order_book(self.market_id)
        bids, asks = order_book['bids'], order_book['asks']
        if side:
            return bids if side == Side.BUY else asks
        return OrderBook(bids=bids, asks=asks)

    def _order_book_entry_amount(self, order):
        return float(order[1])

    def _order_book_entry_price(self, order):
        return float(order[0])


class BitstampWallet(WalletClient, BitstampAuth):

    def _pending_withdraw_amount(self):
        return sum([
            float(w['amount']) for w in self.get_withdrawals()
            if w['status'] in [0]  # status: 0 (open), 1 (in process), 2 (finished), 3 (canceled) or 4 (failed)
        ])

    def _balance(self, currency, available_only=False):
        balances = self.client.account_balance()
        if available_only:
            balance = float(balances[f'{currency.lower()}_available'])
        else:
            # Bitstamp includes withdrawals with status 0 on balance amount
            pending_withdrawal_amount = self._pending_withdraw_amount()
            balance = float(balances[f'{currency.lower()}_balance']) - pending_withdrawal_amount
        return balance

    def _deposits(self, currency: str):
        tx_type = '0'  # '0': deposit; '1': withdrawal; '2': market trade; '14': sub account transfer.
        limit = 1000  # max limit: 1000
        deposits = [d for d in self.client.user_transactions(limit=limit) if d['type'] == tx_type]
        return deposits

    def _withdrawals(self, currency: str):
        time_delta = 50000000  # max delta: 50000000 (seconds)
        withdrawals = self.client.withdrawal_requests(time_delta)
        return [w for w in withdrawals if w['currency'] == currency.upper()]

    def _withdraw(self, currency: str, amount: float, address: str, subtract_fee: bool=False):
        withdrawal_method = {
            'BCH': self.client.bch_withdrawal,
            'BTC': self.client.bitcoin_withdrawal,
            'ETH': self.client.eth_withdrawal,
            'LTC': self.client.litecoin_withdrawal,
            # 'XRP': self.client.xrp_withdrawal,
        }
        # TODO: subtract_fee
        if subtract_fee:
            self.log.warning('subtract_fee option is not implemented!')
        method = withdrawal_method[currency.upper()]
        return method(address, amount)


class BitstampTrading(TradingClient, BitstampAuth, BitstampMarket):
    wallet_client = BitstampWallet
    # TODO: Bitstamp lists min_order size on quote currency https://www.bitstamp.net/api/v2/trading-pairs-info/
    min_order_amount_mapping = {
        'BCH': 0.02,  # ~ 10.00 USD @ 500.00 BTC/USD
        'BTC': 0.002,  # ~ 10.00 USD @ 5000.00 BTC/USD
        'ETH': 0.05,  # ~ 10.00 USD @ 200.00 ETH/USD
        'LTC': 0.2,  # ~ 10.00 USD @ 50.00 LTC/USD
    }

    def _open_orders(self):
        return self.client.open_orders(self.market_id)

    def _order_amount(self, order):
        return float(order['amount'])

    def _cancel_order(self, order):
        return self.client.cancel_order(order['id'])

    def _order_details(self, order_id: int):
        return self.client.orders_status(order_id)

    def _place_order(self, side: Side, o_type: OrderType, amount: float, price: float=None):
        methods = {
            OrderType.MARKET: {
                Side.BUY: self.client.buy_market_order,
                Side.SELL: self.client.sell_market_order,
            },
            OrderType.LIMIT: {
                Side.BUY: self.client.buy_limit_order,
                Side.SELL: self.client.sell_limit_order,
            }
        }
        place_order = methods[o_type][side]
        amount = truncate(amount, 8)
        args = (amount, price) if o_type == OrderType.LIMIT else (amount,)
        return place_order(self.market_id, *args)

    def _open_positions(self):
        return []

    def _open_position(self, side: Side, p_type: OrderType, amount: float, price: float=None, leverage: float=None):
        raise NotImplementedError



import logging
import time

import pygogo

# Set time to UTC
logging.Formatter.converter = time.gmtime

# Set default log format
DEFAULT_CONSOLE_FORMAT = '%(asctime)s UTC | %(name)s | %(levelname)-8s | %(message)s'

# Set formatter
formatter = logging.Formatter(DEFAULT_CONSOLE_FORMAT)

# Initialize pygogo logger
going = pygogo.Gogo(
    name='tradingbots',
    low_formatter=formatter,
    high_formatter=formatter,
    verbose=True,
    monolog=True,
)


def get_logger(name: str='base'):
    return going.get_logger(name)



import os
import shutil
import stat
from importlib import import_module
from os import path

import click
import stringcase
from jinja2 import Template

import trading_bots
from trading_bots.conf import defaults
from trading_bots.conf import settings

# Rewrite the following suffixes when determining the target filename.
rewrite_template_suffixes = (
    # Allow shipping invalid .py files without byte-compilation.
    ('.py.jinja2', '.py'),
)
# The file extension(s) to render
extensions = ('py', 'yml')


def handle_template(bot_or_project, name, target=None, **options):
    """
    Copy either a bot layout template or a Trading-Bots project
    layout template into the specified directory.
    :param bot_or_project: The string 'bot' or 'project'.
    :param name: The name of the bot or project.
    :param target: The directory to which the template should be copied.
    :param options: The additional variables passed to project or bot templates
    """
    bot_or_project = bot_or_project
    paths_to_remove = []
    verbosity = int(options['verbosity'])

    validate_name(name, bot_or_project)

    # if some directory is given, make sure it's nicely expanded
    if target is None:
        top_dir = path.join(os.getcwd(), name)
        try:
            os.makedirs(top_dir)
        except FileExistsError:
            raise click.ClickException("'%s' already exists" % top_dir)
        except OSError as e:
            raise click.ClickException(e)
    else:
        top_dir = os.path.abspath(path.expanduser(target))
        if not os.path.exists(top_dir):
            raise click.ClickException("Destination directory '%s' does not "
                                       "exist, please create it first." % top_dir)

    base_name = '%s_name' % bot_or_project
    base_subdir = '%s_template' % bot_or_project
    base_directory = '%s_directory' % bot_or_project
    target_name = '%s_target' % bot_or_project
    pascal_case_name = 'pascal_case_%s_name' % bot_or_project
    pascal_case_value = stringcase.pascalcase(name)
    snake_case_name = 'snake_case_%s_name' % bot_or_project
    snake_case_value = stringcase.snakecase(name)

    context = {
        **options,
        base_name: name,
        base_directory: top_dir,
        target_name: target,
        pascal_case_name: pascal_case_value,
        snake_case_name: snake_case_value,
        'settings_files': defaults.SETTINGS,
        'version': getattr(trading_bots.__version__, '__version__'),
    }

    # Setup a stub settings environment for template rendering
    settings.configure()
    trading_bots.setup()

    template_dir = path.join(trading_bots.__path__[0], 'conf', base_subdir)
    prefix_length = len(template_dir) + 1

    for root, dirs, files in os.walk(template_dir):

        path_rest = root[prefix_length:]
        relative_dir = path_rest.replace(snake_case_name, snake_case_value)
        if relative_dir:
            target_dir = path.join(top_dir, relative_dir)
            if not path.exists(target_dir):
                os.mkdir(target_dir)

        for dirname in dirs[:]:
            if dirname.startswith('.') or dirname == '__pycache__':
                dirs.remove(dirname)

        for filename in files:
            if filename.endswith(('.pyo', '.pyc', '.py.class')):
                # Ignore some files as they cause various breakages.
                continue
            old_path = path.join(root, filename)
            new_path = path.join(top_dir, relative_dir,
                                 filename.replace(snake_case_name, snake_case_value))
            for old_suffix, new_suffix in rewrite_template_suffixes:
                if new_path.endswith(old_suffix):
                    new_path = new_path[:-len(old_suffix)] + new_suffix
                    break  # Only rewrite once

            if path.exists(new_path):
                raise click.ClickException("%s already exists, overlaying a "
                                           "project or bot into an existing "
                                           "directory won't replace conflicting "
                                           "files" % new_path)

            # Only render the Python files, as we don't want to
            # accidentally render Trading-Bots templates files
            if new_path.endswith(extensions):
                with open(old_path, 'r', encoding='utf-8') as template_file:
                    content = template_file.read()
                template = Template(content, keep_trailing_newline=True)
                content = template.render(**context)
                with open(new_path, 'w', encoding='utf-8') as new_file:
                    new_file.write(content)
            else:
                shutil.copyfile(old_path, new_path)

            if verbosity >= 2:
                click.echo("Creating %s\n" % new_path)
            try:
                shutil.copymode(old_path, new_path)
                make_writeable(new_path)
            except OSError:
                click.echo(
                    "Notice: Couldn't set permission bits on %s. You're "
                    "probably using an uncommon filesystem setup. No "
                    "problem." % new_path)

    if paths_to_remove:
        if verbosity >= 2:
            click.echo("Cleaning up temporary files.\n")
        for path_to_remove in paths_to_remove:
            if path.isfile(path_to_remove):
                os.remove(path_to_remove)
            else:
                shutil.rmtree(path_to_remove)


def validate_name(name, bot_or_project):
    a_or_an = 'an' if bot_or_project == 'bot' else 'a'
    if name is None:
        raise click.ClickException('you must provide {an} {bot} name'.format(
            an=a_or_an,
            bot=bot_or_project,
        ))
    # Check it's a valid directory name.
    if not name.isidentifier():
        raise click.ClickException(
            "'{name}' is not a valid {bot} name. Please make sure the "
            "name is a valid identifier.".format(
                name=name,
                bot=bot_or_project,
            )
        )
    # Check it cannot be imported.
    try:
        import_module(name)
    except ImportError:
        pass
    else:
        raise click.ClickException(
            "'{name}' conflicts with the name of an existing Python "
            "module and cannot be used as {an} {bot} name. Please try "
            "another name.".format(
                name=name,
                an=a_or_an,
                bot=bot_or_project,
            )
        )


def make_writeable(filename):
    """
    Make sure that the file is writeable.
    Useful if our source is read-only.
    """
    if not os.access(filename, os.W_OK):
        st = os.stat(filename)
        new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
        os.chmod(filename, new_permissions)



import logging
from logging import Logger

import pygogo
from pygogo.utils import get_structured_filter

from ..conf import settings

config = settings.logging
bots_root = config['bots_root']
log_level = config['level']
file = settings.logging.get('file')

# Configure logger
console_format = '%(asctime)s UTC | %(tag)s | %(env)-4s | %(bot)s | %(config)s | %(levelname)-8s | %(message)s'
console_formatter = logging.Formatter(console_format)

if file:
    low_level_handler = pygogo.handlers.file_hdlr(file)
else:
    low_level_handler = pygogo.handlers.stdout_hdlr()

gogo = pygogo.Gogo(
    name=config['bots_root'],
    low_formatter=console_formatter,
    high_formatter=console_formatter,
    low_level=log_level['low'],
    high_level=log_level['high'],
    low_hdlr=low_level_handler,
    high_hdlr=pygogo.handlers.stderr_hdlr(),
    monolog=True,
)


def setup_logger(logger: Logger, **logger_kwargs):
    if not logger.handlers:
        logger_name = logger.name.partition('.')[2]
        logger = gogo.get_logger(logger_name)
    structured_filter = get_structured_filter(**logger_kwargs)
    for handler in logger.handlers:
        handler.addFilter(structured_filter)
        handler.setFormatter(console_formatter)
    return logger


def get_logger(logger_name: str):
    logger = logging.getLogger(f'{bots_root}.{logger_name}')
    return logger



from collections import namedtuple
from enum import Enum
from logging import Logger

from cached_property import cached_property

from trading_bots.conf import settings
from trading_bots.core.logging import get_logger
from trading_bots.core.storage import get_store

__all__ = [
    'Market',
    'Side',
    'OrderBook',
    'OrderBookEmptyError',
    'OrderType',
    'BaseClient',
    'CurrencyClientMixin',
    'MarketClientMixin',
    'MarketClient',
    'WalletClient',
    'TradingClient',
]


class Market:

    def __init__(self, code: (str, list, tuple)):
        assert isinstance(code, (str, list, tuple))
        if isinstance(code, (list, tuple)):
            assert len(code) == 2, 'A market code must have 2 items (base and quote currency).'
            base, quote = code
            self.base, self.quote = str(base), str(quote)
        if isinstance(code, str):
            assert len(code) == 6, 'A market code must have 6 characters (base and quote currency).'
            self.base, self.quote = code[:3], code[3:]
        self.code = self.base + self.quote

    def __str__(self):
        return self.code

    def __eq__(self, other):
        if isinstance(other, Market):
            other = other.code
        return self.code == other


OrderBook = namedtuple('order_book', 'bids asks')


class OrderBookEmptyError(Exception):
    pass


class Side(Enum):
    BUY = 'buy'
    SELL = 'sell'


class OrderType(Enum):
    MARKET = 'market'
    LIMIT = 'limit'


class BaseClient:
    name = ''

    def __init__(self, client=None, dry_run: bool=False, timeout: int=None,
                 logger: Logger=None, store=None, **kwargs):
        assert self.name, 'A name must be defined for the client!'
        self.credentials = settings.credentials.get(self.name)
        self.dry_run = dry_run
        self.timeout = timeout
        self.log = logger or get_logger(__name__)
        self.store = store or get_store(self.log)
        self.client = client or self._client()

    def _client(self):
        raise NotImplementedError


class CurrencyClientMixin:

    def __init__(self, currency: str, client=None, dry_run: bool=False, timeout: int=None,
                 logger: Logger=None, store=None, **kwargs):
        super().__init__(client, dry_run, timeout, logger, store, **kwargs)
        self.currency = currency


class MarketClientMixin:

    def __init__(self, market: (str, Market), client=None, dry_run: bool=False, timeout: int=None,
                 logger: Logger=None, store=None, **kwargs):
        super().__init__(client, dry_run, timeout, logger, store, **kwargs)
        if not isinstance(market, Market):
            market = Market(market)
        self.market = market
        self.market_id = self._market_id()

    def _market_id(self):
        return str(self.market)


class WalletClient(CurrencyClientMixin):
    withdrawal_fees = {}

    # Balance
    def _balance(self, currency: str, available_only: bool=False):
        raise NotImplementedError

    def _get_balance(self, currency: str, available_only: bool=False):
        b_type = 'balance' if available_only else 'available amount'
        self.log.debug(f'Obtaining {currency} {b_type} from {self.name}')
        try:
            amount = self._balance(currency, available_only)
            self.log.debug(f'{self.name} {b_type}: {amount} {currency}')
        except Exception:
            self.log.error(f'Failed obtaining {b_type} from {self.name} {currency}')
            raise
        return amount

    def get_balance(self):
        return self._get_balance(self.currency)

    def get_available(self):
        return self._get_balance(self.currency, available_only=True)

    # Deposits
    def _deposits(self, currency: str):
        raise NotImplementedError

    def _get_deposits(self, currency: str):
        self.log.debug(f'Obtaining deposits from {self.name}')
        try:
            deposits = self._deposits(currency)
            self.log.debug(f'Total number of deposits: {len(deposits)}')
        except Exception:
            self.log.error(f'Failed obtaining deposits from {self.name}!')
            raise
        return deposits

    def get_deposits(self):
        return self._get_deposits(self.currency)

    # Withdrawals
    def _withdrawals(self, currency: str):
        raise NotImplementedError

    def _get_withdrawals(self, currency: str):
        self.log.debug(f'Obtaining withdrawals from {self.name}')
        try:
            withdrawals = self._withdrawals(currency)
            self.log.debug(f'Total number of withdrawals: {len(withdrawals)}')
        except Exception:
            self.log.error(f'Failed obtaining withdrawals from {self.name}!')
            raise
        return withdrawals

    def get_withdrawals(self):
        return self._get_withdrawals(self.currency)

    # Request withdrawal
    @cached_property
    def withdrawal_fee(self):
        try:
            return self.withdrawal_fees[self.currency]
        except KeyError:
            return 0

    def _withdraw(self, currency: str, amount: float, address: str, subtract_fee: bool=False):
        raise NotImplementedError

    def _request_withdrawal(self, currency: str, amount: float, address: str, subtract_fee: bool=False):
        withdrawal_msg = self._request_withdrawal_msg(currency)
        self.log.debug(f'Requesting {withdrawal_msg} to {address}')
        # Check dry run, request withdrawal
        if not self.dry_run:
            try:
                withdrawal = self._withdraw(currency, amount, address, subtract_fee)
            except Exception:
                self.log.error(f'Failed requesting {withdrawal_msg}! | Amount: {amount} | Address: {address}')
                raise
            msg = self._withdrawal_details_msg(f'{withdrawal_msg} requested | ', withdrawal)
            self.log.info(msg)
            return withdrawal
        else:
            msg = f'DRY RUN: {withdrawal_msg} requested'
            self.log.warning(msg)
            return False

    def _request_withdrawal_msg(self, currency: str):
        return f'{currency} withdrawal from {self.name}'

    def _withdrawal_details_msg(self, msg: str, withdrawal):
        return f'{msg} {withdrawal}'

    def request_withdrawal(self, amount: float, address: str, subtract_fee: bool=False):
        return self._request_withdrawal(self.currency, amount, address, subtract_fee)


class MarketClient(MarketClientMixin):

    def _ticker(self):
        raise NotImplementedError

    def get_ticker(self):
        self.log.debug(f'Obtaining ticker from {self.name}')
        try:
            ticker = self._ticker()
        except Exception:
            self.log.error(f'Failed obtaining ticker from {self.name}!')
            raise
        return ticker

    def _order_book(self, side: Side=None):
        raise NotImplementedError

    def get_order_book(self, side: Side=None):
        self.log.debug(f'Obtaining order book from {self.name}')
        try:
            order_book = self._order_book(side)
            order_book_len = len(order_book) if side else len(order_book.bids) + len(order_book.asks)
            self.log.debug(f'Order book has {order_book_len} orders')
        except Exception:
            self.log.error(f'Failed obtaining order book from {self.name}!')
            raise
        return order_book

    def _order_book_entry_amount(self, order):
        return order.amount

    def _order_book_entry_price(self, order):
        return order.price

    def _quote_book_price(self, order_book: list, amount: float=0):
        quote = 0
        if not order_book:
            raise OrderBookEmptyError
        for order in order_book:
            quote = self._order_book_entry_price(order)
            amount -= self._order_book_entry_amount(order)
            if amount < 0:
                break
        if amount > 0:
            self.log.warning('Total amount on order book is not enough to cover quote')
        return quote

    def quote_price(self, side: Side, amount: float=0, order_book_side=None):
        # TODO: get price from convert if it fails with  OrderBookEmptyError
        order_book = order_book_side or self.get_order_book(side)
        return self._quote_book_price(order_book, amount)

    def quote_buy_price(self, amount: float=0, order_book_bid=None):
        return self.quote_price(Side.BUY, amount, order_book_bid)

    def quote_sell_price(self, amount: float=0, order_book_ask=None):
        return self.quote_price(Side.SELL, amount, order_book_ask)

    @property
    def min_ask(self):
        return self.quote_buy_price()

    @property
    def max_bid(self):
        return self.quote_sell_price()

    def get_spread_details(self, slippage_amount: float=0, order_book=None):
        order_book = order_book or self.get_order_book()
        max_bid = self.quote_sell_price(slippage_amount, order_book.bids)
        min_ask = self.quote_buy_price(slippage_amount, order_book.asks)
        self.log.debug(f'Market Spread | Bid: {max_bid:10,f} | Ask: {min_ask:10,f}')
        return max_bid, min_ask

    def get_volume_details(self, order_book=None):
        order_book = order_book or self.get_order_book()
        volume_bid = sum([self._order_book_entry_amount(o) for o in order_book.bids])
        volume_ask = sum([self._order_book_entry_amount(o) for o in order_book.asks])
        volume_total = volume_bid + volume_ask
        # Log market volume details
        self.log.debug(f'Bid volume: {volume_bid:10,f}')
        self.log.debug(f'Ask volume: {volume_ask:10,f}')
        self.log.debug(f'Market vol: {volume_total:10,f}')
        if volume_total:
            bid_vol_p = volume_bid / volume_total
            ask_vol_p = volume_ask / volume_total
            self.log.info(f'Market Volume | Bid: {bid_vol_p:6.1%} | Ask: {ask_vol_p:6.1%}')
        else:
            self.log.warning(f'Market has no volume!')
        return volume_bid, volume_ask

    def get_vw_price(self, order_book=None):
        order_book = order_book or self.get_order_book()
        volume_bid, volume_ask = self.get_volume_details(order_book)
        max_bid, min_ask = self.get_spread_details(order_book=order_book)
        volume_total = volume_bid + volume_ask
        if volume_total:
            vw_price = (volume_bid * max_bid + volume_ask * min_ask) / (volume_bid + volume_ask)
            vw_price_str = f'{vw_price:10,f}'
        else:
            vw_price = None
            vw_price_str = 'N/A'
        self.log.info(f'Volume-Weighted Price: {vw_price_str}')
        return vw_price


class TradingClient(MarketClientMixin):
    wallet_client = None
    has_margin_trading = False
    min_order_amount_mapping = {}

    def __init__(self, market, client=None, dry_run: bool=False, timeout: int=None,
                 logger: Logger=None, store=None, **kwargs):
        super().__init__(market, client, dry_run, timeout, logger, store, **kwargs)
        assert self.wallet_client, 'A wallet client must be defined for the client!'
        Wallets = namedtuple('Wallets', 'base quote')
        base = self._wallet_client_init(self.market.base)
        quote = self._wallet_client_init(self.market.quote)
        self.wallets = Wallets(base, quote)

    def _wallet_client_init(self, currency):
        return self.wallet_client(currency, self.client, self.dry_run, self.timeout, self.log)

    # Trading ----------------------------------------------------------------
    def _open_orders(self):
        raise NotImplementedError

    def get_open_orders(self):
        # Log fetch orders message
        self.log.debug(f'Obtaining open orders from {self.name}')
        # Fetch orders
        try:
            orders = self._open_orders()
            self.log.debug(f'Total number of orders: {len(orders)}')
            for order in orders:
                msg = self._order_details_msg(f'Order: ', order)
                self.log.debug(msg)
        except Exception:
            self.log.error(f'Failed obtaining orders from {self.name}!')
            raise
        return orders

    def _order_amount(self, order):
        return order.amount

    def get_open_orders_amount(self):
        orders = self.get_open_orders()
        amount = sum(self._order_amount(o) for o in orders)
        self.log.debug(f'Total amount on orders: {amount}')
        return amount

    def _cancel_order(self, order):
        raise NotImplementedError

    def cancel_order(self, order):
        # Log cancel order message
        self.log.debug('Canceling order')
        # Don't cancel if dry run
        if self.dry_run:
            self.log.warning(f'DRY RUN: Order cancelled')
            return
        # Iterate and cancel orders
        try:
            cancelled_order = self._cancel_order(order)
        except Exception:
            msg = self._order_details_msg('Failed to cancel order: ', order)
            self.log.error(msg)
            raise
        msg = self._order_details_msg('Order cancelled: ', cancelled_order)
        self.log.info(msg)
        return cancelled_order

    def cancel_orders(self, orders: list=None):
        # Log cancel orders message
        self.log.debug('Canceling orders')
        cancelled_orders = []
        # Don't cancel if dry run
        if self.dry_run:
            self.log.warning(f'DRY RUN: Orders cancelled')
            return cancelled_orders
        # Iterate and cancel orders
        orders = orders or self.get_open_orders()
        for order in orders:
            cancelled_order = self.cancel_order(order)
            cancelled_orders.append(cancelled_order)
        self.log.info(f'Cancelled orders: {len(cancelled_orders)}')
        return cancelled_orders

    def _order_details(self, order_id: int):
        raise NotImplementedError

    def order_details(self, order_id: int):
        self.log.debug(f'Obtaining order details from {self.name} for order {order_id}')
        try:
            order = self._order_details(order_id)
        except Exception:
            self.log.error(f'Failed obtaining order details from {self.name}!')
            raise
        msg = self._order_details_msg('Order: ', order)
        self.log.debug(msg)
        return order

    @cached_property
    def min_order_amount(self):
        try:
            return self.min_order_amount_mapping[self.market.base]
        except KeyError:
            return 0

    def _place_order(self, side: Side, o_type: OrderType, amount: float, price: float=None):
        raise NotImplementedError

    def place_order(self, side: Side, o_type: OrderType, amount: float, price: float=None):
        order_msg = self._place_order_msg(side, o_type)
        self.log.debug(f'Placing {order_msg} order')
        # Don´t place order if the amount is lower than the min allowed
        if self.min_order_amount and amount < self.min_order_amount:
            msg = f'{order_msg} order was not placed! {amount:,f} < Min Amount ({self.min_order_amount})'
            self.log.warning(msg)
            return False
        # Check dry run, place order
        if not self.dry_run:
            try:
                new_order = self._place_order(side, o_type, amount, price)
            except Exception:
                self.log.error(f'Failed placing {order_msg} order! | Amount: {amount} | Price: {price}')
                raise
            msg = self._order_details_msg(f'{order_msg} order placed: ', new_order)
            self.log.info(msg)
            return new_order
        else:
            msg = f'DRY RUN: {order_msg} order placed'
            self.log.warning(msg)
            return False

    def place_market_order(self, side: Side, amount: float):
        return self.place_order(side, OrderType.MARKET, amount)

    def place_limit_order(self, side: Side, amount: float, price: float):
        return self.place_order(side, OrderType.LIMIT, amount, price)

    def _place_order_msg(self, side: Side, order_type: OrderType=None):
        msg = side.value
        if order_type:
            msg = f'{msg} {order_type.value}'
        return f'{msg} order'

    def _order_details_msg(self, msg: str, order):
        return f'{msg} {order}'

    # Margin Trading ---------------------------------------------------------
    def _open_positions(self):
        raise NotImplementedError

    def get_open_positions(self):
        # Log fetch positions message
        self.log.debug(f'Obtaining open positions from {self.name}')
        # Fetch positions
        try:
            positions = self._open_positions()
            self.log.debug(f'Total number of positions: {len(positions)}')
        except Exception:
            self.log.error(f'Failed obtaining positions from {self.name}!')
            raise
        return positions

    def _position_amount(self, position):
        return float(position['amount'])

    def get_open_positions_amount(self):
        positions = self.get_open_positions()
        amount = sum(self._position_amount(p) for p in positions)
        self.log.debug(f'Total amount on positions: {amount}')
        return amount

    def _open_position(self, side: Side, o_type: OrderType, amount: float, price: float=None, leverage: float=None):
        raise NotImplementedError

    def open_position(self, side: Side, p_type: OrderType, amount: float, price: float=None, leverage: float=None):
        position_msg = self._open_position_msg(side, p_type)
        self.log.debug(f'Placing {position_msg} order')
        # Don´t place order if the amount is lower than the min allowed
        if self.min_order_amount and amount < self.min_order_amount:
            msg = f'{position_msg} position was not open! {amount:,f} < Min Amount ({self.min_order_amount})'
            self.log.warning(msg)
            return False
        # Check dry run, place position
        if not self.dry_run:
            try:
                new_position = self._open_position(side, p_type, amount, price, leverage)
            except Exception:
                self.log.error(f'Failed to open {position_msg} position!')
                raise
            msg = self._position_details_msg(f'{position_msg} position opened | ', new_position)
            self.log.info(msg)
            return new_position
        else:
            msg = f'DRY RUN: {position_msg} order placed'
            self.log.warning(msg)
            return False

    def open_market_position(self, side: Side, amount: float, leverage: float=None):
        return self.open_position(side, OrderType.MARKET, amount, leverage)

    def open_limit_position(self, side: Side, amount: float, price: float, leverage: float=None):
        return self.open_position(side, OrderType.LIMIT, amount, price, leverage)

    def _open_position_msg(self, side: Side, position_type: OrderType=None):
        msg = side.value
        if position_type:
            msg = f'{msg} {position_type.value}'
        return f'{msg} position'

    def _position_details_msg(self, msg: str, position):
        return f'{msg} {position}'



import os
import sys
import threading
from collections import Counter, OrderedDict, defaultdict
from importlib import import_module
from os import path

from .base import Bot
from ..conf import defaults
from ..core.exceptions import AppRegistryNotReady
from ..core.exceptions import ImproperlyConfigured
from ..core.utils import load_class_by_name
from ..core.utils import load_yaml_file

# The file extension(s) to import as configs
config_extensions = ('py', 'yml')


class BotRegistry:
    """Class representing a Bot and its configuration files."""

    def __init__(self, bot_name, bot_module, label=None):
        # Full Python path to the bot class e.g. 'trading_bots.bots.Bot'.
        self.name = bot_name

        # Root module for the bot e.g. <module 'trading_bots.bots'
        # from 'trading_bots/bots/__init__.py'>.
        self.module = bot_module

        # Bot class for the bot
        self.cls = load_class_by_name(bot_name)

        # Reference to the Bots registry that holds this BotRegistry. Set by the
        # registry when it registers the BotRegistry instance.
        self.bots = None

        # Last component of the Python path to the bot class e.g. 'Bot'.
        # This value must be unique across a project.
        self.label = label or bot_name.rpartition('.')[-1]

        # Filesystem path to the bot directory e.g.
        # '/path/to/trading_bots/bots'.
        self.path = self._path_from_module(bot_module)

        # Filesystem path containing configs e.g. '/path/to/trading_bots/bots'.
        # Set by import_configs().
        # None if the bot doesn't have a configs dir.
        self.configs_path = None

        # Mapping of lower case config file names to config files. Initially set to
        # None to prevent accidental access before import_configs() runs.
        self.configs = OrderedDict()

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.label)

    @staticmethod
    def _path_from_module(module):
        """Attempt to determine bot's filesystem path from its module."""
        # Convert paths to list because Python's _NamespacePath doesn't support
        # indexing.
        paths = list(getattr(module, '__path__', []))
        if len(paths) != 1:
            filename = getattr(module, '__file__', None)
            if filename is not None:
                paths = [os.path.dirname(filename)]
            else:
                # For unknown reasons, sometimes the list returned by __path__
                # contains duplicates that must be removed.
                paths = list(set(paths))
        if len(paths) > 1:
            raise ImproperlyConfigured(
                "The bot module %r has multiple filesystem locations (%r); "
                "you must configure this bot with an AppConfig subclass "
                "with a 'path' class attribute." % (module, paths))
        elif not paths:
            raise ImproperlyConfigured(
                "The bot module %r has no filesystem location, "
                "you must configure this bot with an AppConfig subclass "
                "with a 'path' class attribute." % (module,))
        return paths[0]

    @classmethod
    def create(cls, entry):
        """
        Factory that creates an bot config from an entry in INSTALLED_APPS.
        """
        # trading_bots.example.bot.ExampleBot
        try:
            # If import_module succeeds, entry is a path to a bot module,
            # which may specify a bot class with a default_bot attribute.
            # Otherwise, entry is a path to a bot class or an error.
            module = import_module(entry)

        except ImportError:
            # Track that importing as a bot module failed. If importing as a
            # bot class fails too, we'll trigger the ImportError again.
            module = None

            mod_path, _, cls_name = entry.rpartition('.')

            # Raise the original exception when entry cannot be a path to an
            # bot config class.
            if not mod_path:
                raise

        else:
            try:
                # If this works, the bot module specifies a bot class.
                entry = module.default_bot
            except AttributeError:
                # Otherwise, it simply uses the default bot registry class.
                return cls(f'{entry}.Bot', module)
            else:
                mod_path, _, cls_name = entry.rpartition('.')

        # If we're reaching this point, we must attempt to load the bot
        # class located at <mod_path>.<cls_name>
        mod = import_module(mod_path)
        try:
            bot_cls = getattr(mod, cls_name)
        except AttributeError:
            if module is None:
                # If importing as an bot module failed, that error probably
                # contains the most informative traceback. Trigger it again.
                import_module(entry)
            raise

        # Check for obvious errors. (This check prevents duck typing, but
        # it could be removed if it became a problem in practice.)
        if not issubclass(bot_cls, Bot):
            raise ImproperlyConfigured(
                "'%s' isn't a subclass of Bot." % entry)

        # Entry is a path to an bot config class.
        return cls(entry, mod, bot_cls.label)

    def get_config(self, config_name, require_ready=True):
        """
        Return the config with the given case-insensitive config_name.
        Raise LookupError if no config exists with this name.
        """
        if require_ready:
            self.bots.check_configs_ready()
        else:
            self.bots.check_bots_ready()

        return self.configs.get(config_name.lower(), {})

    def get_configs(self):
        """
        Return an iterable of models.
        """
        self.bots.check_models_ready()
        for config in self.configs.values():
            yield config

    def import_configs(self):
        self.configs_path = os.path.join(self.path, defaults.CONFIGS_DIR_NAME)
        for root, dirs, files in os.walk(self.configs_path):
            for filename in files:
                file_path = path.join(root, filename)
                if file_path.endswith(config_extensions):
                    config = load_yaml_file(file_path)
                    config_name = path.splitext(filename)[0]
                    self.configs[config_name] = config
                    bots.register_config(self.label, config_name, config)


class Bots:
    """
    A registry that stores the installed bots.
    """

    def __init__(self, installed_bots=()):
        # installed_bots is set to None when creating the master registry
        # because it cannot be populated at that point. Other registries must
        # provide a list of installed bots and are populated immediately.
        if installed_bots is None and hasattr(sys.modules[__name__], 'bots'):
            raise RuntimeError("You must supply an installed_bots argument.")

        # Mapping of bot labels => config names => config files.
        self.all_configs = defaultdict(OrderedDict)

        # Mapping of labels to Bot instances for installed bots.
        self.bots = OrderedDict()

        # Whether the registry is populated.
        self.bots_ready = self.configs_ready = self.ready = False

        # Lock for thread-safe population.
        self._lock = threading.RLock()
        self.loading = False

        # Populate bots and models, unless it's the master registry.
        if installed_bots is not None:
            self.populate(installed_bots)

    def populate(self, installed_bots=None):
        """
        Load bots.
        Import each bot module.
        It is thread-safe and idempotent, but not re-entrant.
        """
        if self.ready:
            return

        # populate() might be called by two threads in parallel on servers
        # that create threads before initializing the WSGI callable.
        with self._lock:
            if self.ready:
                return

            # An RLock prevents other threads from entering this section. The
            # compare and set operation below is atomic.
            if self.loading:
                # Prevent re-entrant calls to avoid running AppConfig.ready()
                # methods twice.
                raise RuntimeError("populate() isn't re-entrant")
            self.loading = True

            # Phase 1: Initialize bots
            for entry in installed_bots or {}:
                if isinstance(entry, Bot):
                    cls = entry
                    entry = '.'.join([cls.__module__, cls.__name__])
                bot_reg = BotRegistry.create(entry)
                if bot_reg.label in self.bots:
                    raise ImproperlyConfigured(
                        "Bot labels aren't unique, "
                        "duplicates: %s" % bot_reg.label)

                self.bots[bot_reg.label] = bot_reg
                bot_reg.bots = self

            # Check for duplicate bot names.
            counts = Counter(
                bot_reg.name for bot_reg in self.bots.values())
            duplicates = [
                name for name, count in counts.most_common() if count > 1]
            if duplicates:
                raise ImproperlyConfigured(
                    "Bot names aren't unique, "
                    "duplicates: %s" % ", ".join(duplicates))

            self.bots_ready = True

            # Phase 2: import config files
            for bot in self.bots.values():
                bot.import_configs()

            self.configs_ready = True

            self.ready = True

    def check_bots_ready(self):
        """Raise an exception if all bots haven't been imported yet."""
        if not self.bots_ready:
            raise AppRegistryNotReady("Bots aren't loaded yet.")

    def check_configs_ready(self):
        """Raise an exception if all models haven't been imported yet."""
        if not self.configs_ready:
            raise AppRegistryNotReady("Configs aren't loaded yet.")

    def get_bots(self):
        """Import all bots and return an iterable of bot classes."""
        self.check_bots_ready()
        return self.bots.values()

    def get_bot(self, bot_label):
        """
        Import all bots and returns a bot class for the given label.
        Raise LookupError if no bot exists with this label.
        """
        self.check_bots_ready()
        try:
            return self.bots[bot_label]
        except KeyError:
            message = "No installed bot with label '%s'." % bot_label
            for bot_cls in self.get_bots():
                if bot_cls.name == bot_label:
                    message += " Did you mean '%s'?" % bot_cls.label
                    break
            raise LookupError(message)

    def get_configs(self):
        """
        Return a list of all installed configs.
        """
        self.check_configs_ready()

        result = []
        for bot in self.bots.values():
            result.extend(list(bot.get_models()))
        return result

    def get_config(self, bot_label, config_name=None, require_ready=True):
        """
        Return the config matching the given bot_label and config_name.
        config_name is case-insensitive.
        Raise LookupError if no bot exists with this label, or no config
        exists with this name in the bot. Raise ValueError if called with a
        single argument that doesn't contain exactly one dot.
        """
        if require_ready:
            self.check_configs_ready()
        else:
            self.check_bots_ready()

        if config_name is None:
            config_name = defaults.BOT_CONFIG

        bot = self.get_bot(bot_label)

        if not require_ready and bot.configs is None:
            bot.import_configs()

        return bot.get_config(config_name, require_ready=require_ready)

    def register_config(self, bot_label, config_name, config):
        # Since this method is called when models are imported, it cannot
        # perform imports because of the risk of import loops. It mustn't
        # call get_bot().
        bot_configs = self.all_configs[bot_label]
        bot_configs[config_name] = config

    def is_installed(self, bot_name):
        """
        Check whether a bot with this name exists in the registry.
        bot_name is the full name of the bot e.g. 'trading_bots.core.bot'.
        """
        self.check_bots_ready()
        return any(b.name == bot_name for b in self.bots.values())


bots = Bots(installed_bots=None)



from . import defaults

tag = 'TradingBots'

dry_run = True

installed_bots = ()

logging = {
    'bots_root': defaults.BOTS_LOG_ROOT,
    'filename': defaults.LOG_FILE,
    'level': {
        'low': 'debug',
        'high': 'info',
    },
}

storage = {
    'name': 'json',
    'filename': 'store.json',
}

timeout = 120

urls = {}



from collections import namedtuple

from trading_bots.bots import Bot
from trading_bots.contrib.clients import Market, Side
from trading_bots.contrib.clients import buda
from trading_bots.utils import truncate_to


class RelativeOrders(Bot):
    label = 'RelativeOrders'

    def _setup(self, config):
        # Set market
        self.market = Market(config['market'])
        # Set buda trading client
        self.buda = buda.BudaTrading(
            self.market, dry_run=self.dry_run, timeout=self.timeout, logger=self.log, store=self.store)

    def _algorithm(self):
        # Get middle price from spread
        max_bid, min_ask = self.buda.get_spread_details()
        middle_price = (max_bid + min_ask) / 2
        self.log.info(f'Ticker prices:   Bid: {max_bid} | Ask {min_ask} | Middle {middle_price}')
        # Offset prices from middle using configured price multipliers
        prices_config = self.config['prices']
        bid_price = middle_price * prices_config['buy_multiplier']
        ask_price = middle_price * prices_config['sell_multiplier']
        self.log.info(f'Relative prices: Bid {bid_price} | Ask {ask_price}')
        # Cancel open orders
        self.log.info('Closing open orders')
        self.buda.cancel_orders()

        # Get available amounts
        # Adjust bid and ask amounts
        amounts_config = self.config['amounts']
        amount_base = min(amounts_config['max_base'], self.buda.wallets.base.get_available())
        amount_quote = min(amounts_config['max_quote'], self.buda.wallets.quote.get_available())
        self.log.debug(' | '.join([
            'Amounts',
            f'Bid: {amount_quote} {self.market.quote}',
            f'Ask: {amount_base} {self.market.base}'
        ]))

        # Start strategy
        self.log.info('Starting order deployment')
        # Deploy orders
        deploy_list = self.get_deploy_list()
        self.deploy_orders(deploy_list)

    def _abort(self):
        self.log.error('Aborting strategy, cancelling all orders')
        try:
            self.cancel_orders()
        except Exception:
            self.log.exception(f'Failed!, some orders might not be cancelled')
            raise
        else:
            self.log.info(f'All open orders were cancelled')

    def get_deploy_list(self):

        Order = namedtuple('order', 'amount price side')
        deploy_list = []

        # Available is on quote currency when side is sell
        def quote_to_base_amount(_amount, _price):
            quote_amount = _amount / _price
            return self.truncate_amount(quote_amount)

        buy_order_price = self.truncate_price(self.bid_price)
        buy_order_amount = quote_to_base_amount(self.quote_amount, buy_order_price)
        sell_order_price = self.truncate_price(self.ask_price)
        sell_order_amount = self.truncate_amount(self.base_amount)

        if buy_order_amount > self.buda.min_order_amount:
            deploy_list.append(Order(buy_order_amount, buy_order_price, Side.BUY))
        if sell_order_amount > self.buda.min_order_amount:
            deploy_list.append(Order(sell_order_amount, sell_order_price, Side.SELL))
        return deploy_list

    def deploy_orders(self, deploy_list: list):
        self.log.info(f'Deploying {len(deploy_list)} new orders')
        for order in deploy_list:
            self.buda.place_limit_order(
                side=order.side,
                amount=order.amount,
                price=order.price,
            )

    def truncate_amount(self, value):
        return truncate_to(value, self.market.base)

    def truncate_price(self, value):
        return truncate_to(value, self.market.quote)



import requests

from .base import Converter

__all__ = [
    'BitcoinAverage'
]


class BitcoinAverage(Converter):
    name = 'Bitcoin Average'
    slug = 'bitcoin-average'
    base_url = 'https://apiv2.bitcoinaverage.com'
    symbol_set = 'global'

    def _get_rate(self, currency: str, to: str):
        symbol = (currency + to).upper()
        url = f'{self.base_url}/indices/{self.symbol_set}/ticker/{symbol}'
        response = requests.get(url).json()
        rate = response['last']
        return rate



import time
from logging import Logger

from .logging import get_logger
from .logging import setup_logger
from ..conf import defaults
from ..conf import settings
from ..core.storage import get_store
from ..utils import get_iso_time_str


class Bot:
    """Class representing a base Trading Bot and its logic."""

    label = ''
    verbose_name = ''
    config_file = ''

    def __init__(self, config: dict=None, config_name: str=None, logger: Logger=None):
        assert self.label, 'A Bot object must have a label attribute!'
        # Set configuration
        self.config = config
        self.config_name = config_name or defaults.BOT_CONFIG
        # Configs
        self.dry_run = settings.dry_run
        self.timeout = settings.timeout
        self.env = self.get_env()
        # Set logger
        self.log = logger or self.get_logger()
        self.setup_logger(self.log)
        # Set store
        self.store = get_store(self.log)
        # Time
        self.timestamp = None
        self.run_time = None
        # User setup
        self._setup(self.config)

    def _setup(self, config: dict):
        pass

    def _algorithm(self):
        raise NotImplementedError

    def _abort(self):
        pass

    def _post_exec(self):
        pass

    def execute(self):
        self.timestamp = int(time.time())
        msg = f'Starting {self.label} {self.timestamp}: {get_iso_time_str()} '
        self.log.info(f'{msg:-<80}')

        try:
            self.check_dry_run()
            self._algorithm()

        except Exception:
            self.log.exception(f'Bot entered an invalid state!')
            self._abort()
            raise

        except KeyboardInterrupt:
            self.log.warning(f'Bot execution cancelled!')
            self._abort()
            raise

        finally:
            if self.timestamp:
                self.run_time = time.time() - self.timestamp
                self.log.info(f'Run time: {self.run_time:,.4f} seconds')
                msg = f'Ending {self.label} {self.timestamp}: {get_iso_time_str()} '
            else:
                msg = f'Ending {self.label}: {get_iso_time_str()} '
            self.log.info(f'{msg:-<80}')
            self._post_exec()

    def abort(self):
        try:
            self.log.warning(f'Aborting {self.label} bot...')
            self._abort()
        except Exception:
            self.log.critical(f'Failed to abort!!!', exc_info=True)
            raise

    def setup_logger(self, logger: Logger):
        logger_kwargs = self._logger_kwargs()
        self.log = setup_logger(logger, **logger_kwargs)

    def _get_logger_name(self):
        return f'{self.label}.{self.config_name}'

    def get_logger(self):
        logger_name = self._get_logger_name()
        return get_logger(logger_name)

    def _logger_kwargs(self):
        return {'tag': settings.tag, 'env': self.env, 'bot': self.label, 'config': self.config_name}

    def check_dry_run(self):
        if self.dry_run:
            self.log.warning('DRY RUN!')
        return self.dry_run

    def get_env(self):
        return 'TEST' if self.dry_run else 'LIVE'



from . import __version__


def setup():
    """
    Configure the settings (this happens as a side effect of accessing the
    first setting), configure logging and populate the app registry.
    """
    from trading_bots.bots import bots
    from trading_bots.conf import settings

    bots.populate(settings.installed_bots)



from trading_api_wrappers import Buda

from .base import *

__all__ = [
    'BudaBase',
    'BudaPublic',
    'BudaAuth',
    'BudaMarket',
    'BudaWallet',
    'BudaTrading',
]


class BudaBase(BaseClient):
    name = 'Buda'

    def __init__(self, client=None, dry_run: bool=False, timeout: int=None, logger=None, store=None, host: str=None):
        self.host = host
        super().__init__(client, dry_run, timeout, logger, store)


class BudaPublic(BudaBase):

    def _client(self):
        return Buda.Public(self.timeout, host=self.host)


class BudaAuth(BudaBase):

    def _client(self):
        key = self.credentials['key']
        secret = self.credentials['secret']
        return Buda.Auth(key, secret, timeout=self.timeout, host=self.host)


class BudaMarket(MarketClient, BudaPublic):

    def _market_id(self):
        return f'{self.market.base}-{self.market.quote}'.lower()

    def _ticker(self):
        return self.client.ticker(self.market_id)

    def _order_book(self, side: Side=None):
        order_book = self.client.order_book(self.market_id)
        if side:
            return order_book.bids if side == Side.BUY else order_book.asks
        return order_book


class BudaWallet(WalletClient, BudaAuth):

    def _balance(self, currency, available_only=False):
        balance = self.client.balance(currency)
        if available_only:
            return balance.available_amount.amount
        return balance.amount.amount - balance.pending_withdraw_amount.amount

    @staticmethod
    def _filter_state(items, state: str=None):
        if state:
            items = [i for i in items if i.state == state]
        return items

    def _deposits(self, currency: str):
        return self.client.deposits(currency)

    def _withdrawals(self, currency: str):
        return self.client.withdrawals(currency)

    def _withdraw(self, currency: str, amount: float, address: str, subtract_fee: bool=False):
        if self.dry_run:
            return self.client.simulate_withdrawal(currency, amount, amount_includes_fee=subtract_fee)
        return self.client.withdrawal(currency, amount, address, amount_includes_fee=subtract_fee)

    def _withdrawal_details_msg(self, msg: str, withdrawal):
        # Limit is null on a market order
        msg += ('ID: {w.id} | '
                'Amount: {w.amount.amount:,f} | '
                'Fee: {w.fee.amount:,f} | '
                'Address: {w.data.address} | '
                'State: {w.state}')
        return msg.format(w=withdrawal)


class BudaTrading(TradingClient, BudaAuth, BudaMarket):
    wallet_client = BudaWallet
    side_mapping = {
        Side.BUY: Buda.OrderType.BID,
        Side.SELL: Buda.OrderType.ASK,
    }
    min_order_amount_mapping = {
        'BCH': 0.0001,
        'BTC': 0.0001,
        'ETH': 0.001,
        'LTC': 0.00001,
    }

    def _open_orders(self):
        return self.client.order_pages(self.market_id, state=Buda.OrderState.PENDING).orders

    def _order_amount(self, order):
        return order.amount.amount

    def _cancel_order(self, order):
        return self.client.cancel_order(order.id)

    def _order_details(self, order_id: int):
        return self.client.order_details(order_id=order_id)

    def _place_order(self, side: Side, o_type: OrderType, amount: float, price: float=None):
        side = self.side_mapping[side].value
        return self.client.new_order(self.market_id, side, o_type.value, amount, price)

    def _place_order_msg(self, side: Side, order_type: OrderType=None):
        return f'{side.value.title()} {order_type.value.title()}'

    def _order_details_msg(self, msg: str, order):
        # Limit is null on a market order
        price_type = Buda.OrderPriceType.LIMIT.value
        limit = order.limit.amount if order.price_type == price_type else 0
        msg += ('ID: {o.id} | '
                'Type: {o.type} | '
                'Limit: {limit:,f} | '
                'Amount: {o.amount.amount:,f} | '
                'State: {o.state}')
        return msg.format(o=order, limit=limit)

    def _open_positions(self):
        raise NotImplementedError

    def _open_position(self, side: Side, p_type: OrderType, amount: float, price: float=None, leverage: float=None):
        raise NotImplementedError



import os

import click

from trading_bots.bots import BotTask
from trading_bots.conf import defaults
from .templates import handle_template

# Parameters
bot_argument = click.argument('bot')

config_option = click.option(
    '--config', '-c',
    default=defaults.BOT_CONFIG,
    help="Specify your Bot's configuration filename (YAML format). "
         "By default, configuration files must be placed inside the Bot's 'configs/' folder. "
         f"Defaults to '{defaults.BOT_CONFIG}'.")

interval_option = click.option(
    '--interval', '-i',
    default=defaults.LOOP_INTERVAL,
    help="Loop interval (in seconds). "
         f"Defaults to {defaults.LOOP_INTERVAL}s")

settings_option = click.option(
    '--settings',
    help="Specify your global settings files (YAML format), "
         f"e.g '{defaults.SETTINGS}'.")


verbosity_option = click.option(
    '-v', '--verbosity',
    default=str(defaults.VERBOSITY),
    type=click.Choice(['0', '1', '2', '3']),
    help="Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output")


def print_banner():
    click.echo()
    click.echo('TRADING BOTS 🤖')
    click.echo('===============')
    click.echo()


def print_options(bot, config, settings):
    click.echo(f'Global settings')
    click.echo(f"- Settings files: {os.environ.get('SETTINGS', settings)}")
    click.echo()
    click.echo(f'Bot: {bot}')
    click.echo(f"- Config file: {config or 'default'}")


# CLI
@click.group()
def cli():
    print_banner()


@cli.command(short_help="Execute a bot's logic")
@bot_argument
@config_option
@settings_option
def run(bot, config, settings):
    """Run a specified BOT by label e.g. 'MyBot'"""
    print_options(bot, config, settings)
    click.echo()
    bot_task = BotTask(bot, config)
    bot_task.run_once()


@cli.command(short_help="Schedule a bot to run on an interval")
@bot_argument
@config_option
@interval_option
@settings_option
def loop(bot, config, interval, settings):
    """Schedule a BOT (by label) to run on an interval, e.g. 'MyBot -i 60'"""
    print_options(bot, config, settings)
    click.echo(f'- Interval: {interval}s')
    click.echo()
    bot_task = BotTask(bot, config)
    bot_task.run_loop(interval)


@cli.command(short_help="Execute a bot's abort command")
@bot_argument
@config_option
@settings_option
def abort(bot, config, settings):
    """Run the abort command of a specified BOT by label e.g. 'MyBot'"""
    print_options(bot, config, settings)
    click.echo()
    bot_task = BotTask(bot, config)
    bot_task.abort()


@cli.command(short_help="Creates a Trading-Bots project directory structure")
@click.option('--name', prompt=True, default='MyAwesomeProject')
@click.option('--directory', prompt=True, default='.')
@verbosity_option
def startproject(name, directory, verbosity):
    """
    Creates a Trading-Bots project directory structure for the given project
    NAME in the current directory or optionally in the given DIRECTORY.
    """
    handle_template('project', name, target=directory, verbosity=verbosity)
    click.echo(f"Success: '{name}' project was successfully created on '{directory}'")


@cli.command(short_help="Creates a Bot's directory structure")
@click.option('--name', prompt=True, default='MyAwesomeBot')
@click.option('--directory', prompt="Directory (your project's dir)")
@verbosity_option
def createbot(name, directory, verbosity):
    """
    Creates a Bot's directory structure for the given bot NAME in
    the current directory or optionally in the given DIRECTORY.
    """
    handle_template('bot', name, target=directory, verbosity=verbosity)
    click.echo(f"Success: '{name}' bot was successfully created on '{directory}'")
